/*
  ulistd - FBB unproto lists recieving daemon
  (c) 1998 - 1999 by Radek Burget OK2JBG (xburge01@stud.fee.vutbr.cz)

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version
  2 of the license, or (at your option) any later version.

  HISTORY - started at version 0.08

  30 Nov 99 - version 0.09 - the index file is not read again and again
                             to get the actual number of messages

  14 Jan 00 - version 0.10 - SSID is ignored in the path to list files

*/

#include <ctype.h>
#include <fcntl.h>
#include <stdio.h>
#include <signal.h>
#include <string.h>
#include <stdlib.h>
#include <syslog.h>
#include <sys/time.h>
#include <unistd.h>
#include <netinet/in.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <linux/ax25.h>
#include <linux/if_ether.h>

#include "calls.h"
#include "daemon.h"

#define CALL_OFS 7
#define DATA_OFS 16
#define MIN_SIZE 15

#define AXPORTS "/etc/ax25/axports"
#define CONFIG "/etc/ax25/ulistd.conf"
#define OUTDIR "/var/ax25/ulistd"
#define EXECCMD "/var/ax25/list_agent"
#define MAX_PORT 32
#define MAX_BBS 16

/* how often the frames are read (seconds) */
#define SCAN_TIME 1

/* default time to wait for response (seconds) */
#define RESP_TIME 15

/* default time to wait for next frame (seconds) */
#define WAIT_TIME 10

/* number of unsuccesful tries before bailing out */
#define MAX_TRIES 32

#define DAEMON

#ifndef VERSION
#define VERSION "0.10"
#endif

/* port info - for each port */
char *pname[MAX_PORT];       /* port names */
char *bcall[MAX_PORT];       /* base calls */
/* bbs info  - for each bbs */
char *bpath[MAX_BBS];        /* bbs paths */
char *bport[MAX_BBS];        /* bbs ports */
int enabled[MAX_BBS];        /* sending is enabled */
time_t acktime[MAX_BBS];     /* time of last reaction of bbs */
time_t rqtime[MAX_BBS];      /* time of last request sent */
int resptime[MAX_BBS];       /* how long to wait for response (sec) */
int waittime[MAX_BBS];       /* how long to wait for next frame (sec) */
int lastmsg[MAX_BBS];        /* the highest msg number heard */
int atend[MAX_BBS];          /* !0 = all messages were received */
int check_end[MAX_BBS];      /* check the end of list (initial download) */
int tries[MAX_BBS];          /* number of tries */
int last[MAX_BBS];           /* last message number in the index file */

/* setup */
char mycall[20];             /* my callsign */
int num_ports;               /* number of ports */
int num_bbs;                 /* number of bbss */
int req_num;                 /* requested msg number -1 = none */
int logging = 0;             /* write to syslog ? */
int startreq = 1;            /* send requests at starup ? */
int minmsg = 1;              /* start asking this number */
int getmail = 0;             /* start "user agent" on end of list */

/* flags */
int sleeping = 0;            /* 1=no frames recently - sleeping */

/*======================== Callsign transformations =======================*/

int call_ssid(const char *src)
{
  char *p = (char *)(src);
  while (*p != '\0' && *p != '-') p++;
  if (*p) p++;
  if (*p) return atoi(p);
     else return 0;
}

char *call_call(const char *src)
{
  static char s[15];
  char *p = (char *)(src);
  strcpy(s, "");
  while (*p && isalnum(*p) && strlen(s) < 6) {strncat(s, p, 1); p++;}
  return s;
}

char *strip_ssid(char *src)
{
  if (call_ssid(src) == 0)
  {
    char *p = src;
    while (*p != '\0' && *p != '-') p++;
    *p = '\0';
  }
  return src;
}

char *normalize_call(char *call)
{
  char c[8];
  int ssid;
  char *p;

  strcpy(c, call_call(call));
  ssid = call_ssid(call);
  if (ssid == 0) strcpy(call, c); else sprintf(call, "%s-%i", c, ssid);
  for (p = call; *p; p++) *p = toupper(*p);
  return call;
}

/* Compare callsign only (not the SSID) */
int compare_call(char *call1, char *call2)
{
  char c1[15];
  char c2[15];
  strcpy(c1, call_call(call1));
  strcpy(c2, call_call(call2));
  return (strcasecmp(c1, c2) == 0);
}

/* return the first callsign in the path */
char *first_call(const char *src, char *dest)
{
  char *s = (char *)(src);
  char *d = dest;
  while (isalnum(*s) || *s == '-') *d++ = *s++;
  *d = '\0';
  return dest;
}

/* read axports file */
int read_axports()
{
  char buf[256];
  FILE *f;
  int i=0;
  int speed, paclen, win;
  char descript[256];

  f = fopen(AXPORTS, "r");
  if (f==NULL) return -1;
  while (!feof(f))
  {
    strcpy(buf, "");
    fgets(buf, 255, f);
    if (strlen(buf) != 0 && buf[0] != '#')
    {
      pname[i] = (char *) malloc(sizeof(char) * 16);
      bcall[i] = (char *) malloc(sizeof(char) * 10);
      sscanf(buf, "%s %s %i %i %i %s", pname[i], bcall[i], &speed,
                                      &paclen, &win, descript);
      i++;
    }
  }
  fclose(f);
  return i;
}

/* read configuration file */
int read_config()
{
  FILE *config;
  char s[256];
  char *path, *endptr, *ports;
  int i;

  config = fopen(CONFIG, "r");
  if (config == NULL) return 0;
  
  i = 0;
  while (!feof(config))
  {
    strcpy(s, "");
    fgets(s, 255, config);
    if (s[0] != '#' && strlen(s) > 0)
    {
      path = s;

      //read timings
      resptime[i] = strtol(path, &endptr, 10); path = endptr;
      waittime[i] = strtol(path, &endptr, 10); path = endptr;

      //read port
      while (isspace(*path) && *path) path++;
      ports = path;
      
      //read path
      while (!isspace(*path) && *path) path++; //skip port name
      *path = '\0'; path++;
      while (isspace(*path) && *path) path++;  //start of path

      bport[i] = (char *) malloc((strlen(ports)+1)*sizeof(char));
      bpath[i] = (char *) malloc((strlen(path)+1)*sizeof(char));
      strcpy(bport[i], ports);
      strcpy(bpath[i], path);

      i++;
      if (i >= MAX_BBS) break;
    }
  }
  fclose(config);
  return i;
}

/* return the index of the BBS */
int bbs_num(char *bbs)
{
  int i;
  char path[256];
  
  for (i=0; i<num_bbs; i++)
  {
    first_call(bpath[i], path);
    normalize_call(path);
    if (strcasecmp(path, bbs) == 0) return i;
  }
  return -1;
}

/*=========================== Mail list operations ========================*/

/* returns the number of last message on the list */
int last_message(int bbsnum)
{
  int msgnum = minmsg;

  if (last[bbsnum] == -1) //not read from the file yet
  {
    FILE *f;
    char s[256];
    int i, n;
    char bbs[256];
    
    first_call(bpath[bbsnum], bbs);
    normalize_call(bbs);
  
    sprintf(s, "%s/%s", OUTDIR, call_call(bbs));
    f = fopen(s, "r");
    if (f != NULL)
    {
      while (!feof(f))
      {
        fgets(s, 255, f);
        n = sscanf(s, "%i", &i);
        if (n >= 1) msgnum = i;
      }
      fclose(f);
    }
    else
    {
      f = fopen(s, "w");
      if (f == NULL)
      {
        if (logging)
          syslog(LOG_ERR, "Cannot create file %s", s);
      }
      else fclose(f);
    }
    last[bbsnum] = msgnum;
  }
  else
    msgnum = last[bbsnum];

  return msgnum;
}

/* add next line to list */
void add_line(char *bbs, int bbsnum, char *line)
{
  FILE *f;
  char s[256];

  sprintf(s, "%s/%s", OUTDIR, call_call(bbs));
  f = fopen(s, "a");
  if (f == NULL && logging)
  {
    syslog(LOG_ERR, "Cannot write to file %s\n", s);
  }

  fprintf(f, "%s\n", line);
  last[bbsnum]++;
  fclose(f);
}

/* disable sending requests to this BBS */
void disable_bbs(char *bbs)
{
  int i;
  i = bbs_num(bbs);
  if (i != -1) enabled[i] = 0;
}

/*============================ BBS Communication ==========================*/

void exec_agent(char *bbs)
{
  char cmd[1024];
  sprintf(cmd, "%s %s >/dev/null", EXECCMD, call_call(bbs));
#ifndef DAEMON
  printf("Executing: %s", cmd);
#endif
  system(cmd);
}

void send_request(char *bbs, int num)
{
  int s;
  int i, port, pnum, sum, len, bbsnum;
  char path[30];
  char msg[30];
  char *p;
  char *dest;
  struct full_sockaddr_ax25 addr;

#ifndef DAEMON
  printf("Request to %s\n", bbs);
#endif

  /* find port name */
  p = NULL;
  i = bbs_num(bbs);
  if (i != -1)
  {
    if (!enabled[i]) return; //requests not enabled
    p = bport[i];
    dest = bpath[i];
  }

  bbsnum = i;

  if (p == NULL)
  {
    return; /* unknown BBS */
  }

  /* find interface callsign of the port */
  port = -1;
  for (i = 0; i < num_ports; i++)
    if (strcmp(pname[i], p) == 0) {port = i; break;}

  if (port == -1)
  {
    if (logging)
      syslog(LOG_ERR, "Unknown port %s in %s", p, AXPORTS);
    return;
  }

  /* calculate request */
  pnum = num;
  sum = 0;
  while (pnum > 0)
  {
    sum += (pnum % 256);
    pnum /= 256;
  }
  sprintf(msg, "? %08X%02X", num, sum & 0xFF);

  /* send unproto frame */
  if ((s = socket(AF_AX25, SOCK_DGRAM, 0)) == -1)
    if (logging)
      syslog(LOG_ERR, "Cannot create socket for unproto");
  sprintf(path, "%s %s", mycall, bcall[port]);
  len = convert_path(path, &addr);
  if (len == -1)
    if (logging)
      syslog(LOG_ERR, "Invalid path %s %s", mycall, bcall[port]);
  if (bind(s, (struct sockaddr *)&addr, len) == -1)
     if (logging)
       syslog(LOG_ERR, "Cannot bind() socket for unproto");
  len = convert_path(dest, &addr);
  if (len == -1)
    if (logging)
      syslog(LOG_ERR, "Invalid path %s %s", dest, bcall[port]);
  if (sendto(s, msg, strlen(msg), 0, (struct sockaddr *)&addr, len) == -1)
     if (logging)
       syslog(LOG_ERR, "Cannot sendto() unproto message");
  close(s);

  rqtime[bbsnum] = time(NULL);
  acktime[bbsnum] = 0;

  tries[bbsnum]++;
  if (tries[bbsnum] > MAX_TRIES)
  {
    tries[bbsnum] = 0;
    atend[bbsnum] = 1; //stop sending requests
    if (logging)
      syslog(LOG_ERR, "Specified BBS `%s' is not responding", bbs);
  }
}

/* handle one unproto data line */
void handle_fbb_line(char *bbs, char *line)
{
  int msgnum;
  int lastnum;
  int bbsnum;
  int n;
  char s1[256], s2[256];

  bbsnum = bbs_num(bbs);
  if (bbsnum == -1) return; //some other BBS
  
  lastnum = last_message(bbsnum);
                     
  n = sscanf(line, "%i %s %s", &msgnum, s1, s2);
  if (n > 0)
  {
    acktime[bbsnum] = time(NULL); //BBS has sent s response
    rqtime[bbsnum] = 0;
    tries[bbsnum] = 0;
    if (msgnum > lastmsg[bbsnum]) lastmsg[bbsnum] = msgnum; //last msg heard
    if (strcmp(s1, "/") == 0) //request rejected
    {
      if (compare_call(s2, mycall))
      {
        disable_bbs(bbs);
        if (logging)
          syslog(LOG_ERR, "Request rejected by %s", bbs);
      }
    }
    else if (strcmp(s1, "!") == 0) //change list start
    {
      if (compare_call(s2, mycall))
      {
        char s[20];
        sprintf(s, "%i #", msgnum);
        add_line(bbs, bbsnum, s); last[bbsnum] = msgnum;
      }
    }
    else if (strcmp(s1, "!!") == 0) //end of list
    {
      if (msgnum > lastnum + 1) //incomplete list
      {
        req_num = lastnum;
        atend[bbsnum] = 0;
        check_end[bbsnum] = 1; //list is incomplete, check the end
      }
      else //list OK
      {
        atend[bbsnum] = 1;
        check_end[bbsnum] = 0; //do not check ends more, list should be complete
        if (getmail) exec_agent(bbs); //run agent
      }
    }
    else //message line
    {
      if (msgnum == lastnum + 1) //fits to the list
      {
        add_line(bbs, bbsnum, line);
        atend[bbsnum] = 0;   //possibly it's not a last one
      }
      else if (msgnum > lastnum + 1) //a hole in the list
      {
        req_num = lastnum;     //request from our last number
        check_end[bbsnum] = 1; //list is incomplete, check it
      }
    }
  }
}

/* send initial requests to all BBSs */
void send_init_requests()
{
  char bbs[256];
  int i;

  for (i=0; i<num_bbs; i++)
  {
    first_call(bpath[i], bbs);
    normalize_call(bbs);
    send_request(bbs, last_message(i));
  }
}

/*=========================================================================*/

/* check if all messages have been received */
void check()
{
  int i;
  time_t now;
  int lastnum;
  char bbs[20];

  now = time(NULL);
  /*printf("check %li %li %li\n", now, rqtime[0]+RESP_TIME, acktime[0]+WAIT_TIME);*/
  for (i=0; i<num_bbs; i++)
    if ((rqtime[i] != 0 && now >= rqtime[i] + resptime[i]) ||
        (acktime[i] != 0 && now >= acktime[i] + waittime[i] && !atend[i]))
    {
      lastnum = last_message(i);
      first_call(bpath[i], bbs);
      normalize_call(bbs);
      /*printf("request %s %i atend: %i\n", bbs, lastnum, atend[i]);*/
      if ((!atend[i] && check_end[i]) //surely not on the end
           || lastnum < lastmsg[i])   //or we heard a highest number
        send_request(bbs, lastnum);   //send the request
    }
}

/* TERM signal handler */
void terminate(int sig)
{
  exit(0);
}


int main(int argc, char **argv)
{
  int s; /* socket */
  char buffer[1024];
  char *text;
  struct sockaddr sa;
  int size;
  int textlen;
  int i;

  char from_call[20];
  char to_call[20];

  strcpy(mycall, "");
  for (i=0; i<MAX_BBS; i++)
  {
    enabled[i] = 1;
    rqtime[i] = 0;
    acktime[i] = 0;
    resptime[i] = RESP_TIME;
    waittime[i] = WAIT_TIME;
    atend[i] = 0;
    check_end[i] = 1;
    tries[i] = 0;
    last[i] = -1;
  }

  while ((i = getopt(argc, argv, "c:nlgm:vh")) != -1)
  {
     switch (i)
     {
       case 'c': strncpy(mycall, optarg, 15); break;
       case 'n': startreq = 0; break;
       case 'l': logging = 1; break;
       case 'g': getmail = 1; break;
       case 'm': minmsg = atoi(optarg); break;
       case 'v': printf("%s\n", VERSION); return 0; break;
       case 'h':
       case ':':
       case '?': printf("ulistd daemon by OK2JBG\n");
                 printf("usage: ulistd -c <mycall> [-n] [-l] [-g] [-m number] [-v] [-h]\n");
                 return 1;
                 break;
     }
  }

  normalize_call(mycall);

  if (strlen(mycall) == 0)
  {
    fprintf(stderr, "Callsign is not set. Use -c option.\n");
    return 1;
  }

  num_bbs = read_config();
  if (num_bbs <= 0)
  {
    fprintf(stderr, "no BBS specified in configuration file %s\n", CONFIG);
    return 1;
  }

  num_ports = read_axports();
  if (num_ports == -1)
  {
    fprintf(stderr, "cannot read file %s\n", AXPORTS);
    return 1;
  }

  #ifndef DAEMON
  printf("Reqtime=%i, Waittime=%i\n", resptime[0], waittime[0]);
  #endif

  signal(SIGTERM, terminate);

  if ((s = socket(AF_INET, SOCK_PACKET, htons(ETH_P_AX25))) == -1)
  {
    perror("ulistd: socket");
  }
  fcntl(s, F_SETFL, O_NONBLOCK);

#ifdef DAEMON
  if (!daemon_start(0))
  {
    fprintf(stderr, "ulistd: cannot become a daemon\n");
    return 1;
  }
#endif

  if (logging)
    openlog("ulistd", LOG_PID, LOG_DAEMON);

  if (startreq)
    send_init_requests();

  while (1)
  {
    int asize = sizeof(sa);
    if (sleeping) sleep(SCAN_TIME);
    size = recvfrom(s, buffer, sizeof(buffer), 0, &sa, &asize);
    if (size != -1 && sa.sa_family == AF_AX25 && size >= MIN_SIZE)
    {
      signed char *data = buffer; data++;
      sleeping = 0;
      norm_call(to_call, (ax25_address *)data);
      norm_call(from_call, (ax25_address *)(data + CALL_OFS));
      normalize_call(to_call);
      normalize_call(from_call);
      #ifndef DAEMON
      printf("From: %s  To: %s  Data: %i\n", from_call, to_call, size);
      #endif
      if (size > DATA_OFS)
      {
        int poz = DATA_OFS - 1;
        while (data[poz] != -16 && poz < size-1) poz++; //find start of data
        poz++;
        text = data + poz;
        textlen = size - poz - 1;
        text[textlen] = '\0';

        #ifndef DAEMON
        printf("Text(%i): %s\n", textlen, text);
        #endif
      }

      /* handle FBB unproto frame */
      if (strcmp(to_call, "FBB") == 0)
      {
        char *p = text;
        char line[256];

        req_num = -1;
        strcpy(line, "");
        while (*p)
        {
          if (*p == '\r' || *p == '\n')
          {
            if (strlen(line) > 0) handle_fbb_line(from_call, line);
            strcpy(line, "");
          }
          else
            strncat(line, p, 1);
          p++;
        }
        if (strlen(line) > 0) handle_fbb_line(from_call, line);
        if (req_num > -1) send_request(from_call, req_num);
      }
    } else sleeping = 1; /* no frame received */
    check();
  }
}

